@@ -1,19 +1,17 @@
 import ast
-from dataclasses import dataclass, field
 import enum
 import inspect
 import re
 import types
-import typing as t
+from dataclasses import dataclass, field
 from functools import reduce
+from typing import Any, List, Dict, Optional, Union, Callable
 
 import jsbeautifier
 from strbuilder import BaseBuilder, SurroundBuilder
 
 from .visitor import Visitor, VisitorContext
 
-transformers: t.List['Transformer'] = []
-
 
 class JSScope(enum.Enum):
     MODULE = 0
@@ -23,73 +21,33 @@ class JSScope(enum.Enum):
     F_STRING = 4
     FOR_LOOP = 5
 
-BOOL_OP_TABLE = {
-    ast.Or: '||',
-    ast.And: '&&'
-}
-UNARY_OP_TABLE = {
-    ast.Invert: '~',
-    ast.Not: '!',
-    ast.UAdd: '+',
-    ast.USub: '-',
-}
-COMPARE_OP_TABLE = {
-    ast.Eq: '==',
-    ast.Is: '===',
-    ast.Gt: '>',
-    ast.Lt: '<',
-    ast.IsNot: '!==',
-    ast.NotIn: 'in',
-    ast.In: 'in'
-}
-OPERATOR_OP_TABLE = {
-    ast.Add: '+',
-    ast.BitAnd: '&',
-    ast.BitOr: '|',
-    ast.BitXor: '^',
-    ast.Div: '/',
-    ast.FloorDiv: '//',
-    ast.LShift: '<<',
-    ast.Mod: '%',
-    ast.Mult: '*',
-    ast.Pow: '**',
-    ast.RShift: '>>',
-    ast.Sub: '-',
-}
-CONSTANT_TABLE = {
-}
-
 
 @dataclass
 class JSVisitorContext(VisitorContext):
-    bool_op_table: t.Dict[type, str] = field(default_factory=BOOL_OP_TABLE.copy)
-    unary_op_table: t.Dict[type, str] = field(default_factory=UNARY_OP_TABLE.copy)
-    compare_op_table: t.Dict[t.Any, str] = field(default_factory=COMPARE_OP_TABLE.copy)
-    operator_op_table: t.Dict[type, str] = field(default_factory=OPERATOR_OP_TABLE.copy)
-    constant_table: t.Dict[t.Any, str] = field(default_factory=CONSTANT_TABLE.copy)
-    variables: t.List[str] = field(default_factory=list)
+    variables: List[str] = field(default_factory=list)
+    global_variables: list[str] = field(default_factory=list)
     scope: JSScope = field(default=JSScope.MODULE)
     break_suffix: str = field(default='')
     parent: ast.AST = field(default=None)
     node: ast.AST = field(default=None)
     has_yield: bool = field(default=False)
 
+    def is_existed(self, name: str) -> bool:
+        return name in self.variables or name in self.global_variables
+
     def copy(
         self,
-        variables: t.Dict[str, t.Any] = None,
-        node: t.Optional[ast.AST] = None,
-        scope: t.Optional[JSScope] = None,
-        break_suffix: t.Optional[str] = None,
-        has_yield: t.Optional[bool] = None
+        variables: Dict[str, Any] = None,
+        node: Optional[ast.AST] = None,
+        scope: Optional[JSScope] = None,
+        break_suffix: Optional[str] = None,
+        has_yield: Optional[bool] = None,
+        root: Optional['JSVisitorContext'] = None
     ) -> 'JSVisitorContext':
         return JSVisitorContext(
-            bool_op_table=self.bool_op_table,
-            unary_op_table=self.unary_op_table,
-            compare_op_table=self.compare_op_table,
-            operator_op_table=self.operator_op_table,
             variables=variables or self.variables,
+            global_variables=self.global_variables,
             scope=scope or self.scope,
-            constant_table=self.constant_table,
             break_suffix=self.break_suffix if break_suffix is None else break_suffix,
             parent=self.node,
             node=node or self.node,
@@ -97,13 +55,65 @@ class JSVisitorContext(VisitorContext):
         )
 
 
+class ConstantConsumer:
+    def __call__(self, value: Any, ctx: JSVisitorContext) -> str: ...
+
+
 class CodeGen(Visitor[JSVisitorContext]):
+    BOOL_OP_TABLE: Dict[type, str] = {
+        ast.Or: '||',
+        ast.And: '&&'
+    }
+
+    UNARY_OP_TABLE: Dict[type, str] = {
+        ast.Invert: '~',
+        ast.Not: '!',
+        ast.UAdd: '+',
+        ast.USub: '-',
+    }
+
+    COMPARE_OP_TABLE: Dict[type, str] = {
+        ast.Eq: '==',
+        ast.NotEq: '!==',
+        ast.Is: '===',
+        ast.Gt: '>',
+        ast.GtE: '>=',
+        ast.Lt: '<',
+        ast.LtE: '<=',
+        ast.IsNot: '!==',
+        ast.NotIn: 'in',
+        ast.In: 'in'
+    }
+
+    OPERATOR_OP_TABLE: Dict[type, str] = {
+        ast.Add: '+',
+        ast.BitAnd: '&',
+        ast.BitOr: '|',
+        ast.BitXor: '^',
+        ast.Div: '/',
+        ast.FloorDiv: '/',
+        ast.LShift: '<<',
+        ast.Mod: '%',
+        ast.Mult: '*',
+        ast.Pow: '**',
+        ast.RShift: '>>',
+        ast.Sub: '-',
+    }
+
+    CONSTANT_CONSUMER_TABLE: Dict[type, ConstantConsumer] = {
+        bool: lambda value, ctx: 'true' if value else 'false',
+        str: lambda value, ctx: ('%s' if ctx.scope == JSScope.F_STRING else f'`%s`') % value.replace('`', '\\`').replace('$', '\\$'),
+        int: lambda value, ctx: str(value),
+        float: lambda value, ctx: str(value),
+        type(None): lambda _, ctx: 'null',
+        type(...): lambda _, ctx: 'ellipsis'
+    }
 
     def __init__(self, python_compatible: bool = False) -> None:
         self.python_compatible = python_compatible
         super().__init__(JSVisitorContext)
 
-    def sperator(self, ast: t.Union[t.List[ast.AST], ast.AST], context: t.Optional[JSVisitorContext] = None):
+    def sperator(self, ast: Union[List[ast.AST], ast.AST], context: Optional[JSVisitorContext] = None):
         if context.scope == JSScope.ARGUMENT:
             return ','
         return ';'
@@ -126,21 +136,22 @@ class CodeGen(Visitor[JSVisitorContext]):
             .write('from')\
             .write(SurroundBuilder(node.module))\
             .build()
-    
+
     # def visit_Comment(self, node: ast.PyCF_TYPE_COMMENTS)
 
     def visit_ClassDef(self, node: ast.ClassDef, ctx: JSVisitorContext):
         context = ctx.copy(scope=JSScope.CLASS_FIELD)
 
         body = node.body
-
+        name = ast.Name(node.name)
         builder = BaseBuilder()\
             .write(f'let {node.name} = class')\
             .write(SurroundBuilder(surround='{}')
-                   .write_if(self.python_compatible, '''constructor(...args) {if ('__init__' in this) this.__init__(this, ...args); return new Proxy(this, { apply: (target, self, args) => target.__call__(self, ...args), get: (target, prop, receiver) => {if (target[prop] instanceof Function) {return (...args) => target[prop](target, ...args)} else {return target[prop]}}})}''')
-                   .write(self.visit(node.body, context))
-                   .write(self.visit(filter(lambda node: not isinstance(node, ast.FunctionDef), body), context)))\
-            .write_if(node.bases, (lambda: f'''Object.getOwnPropertyNames({base}.prototype).forEach(name => {{if (name !== 'constructor') {{{node.name}.prototype[name] = {base}.prototype[name];}}}});''' for base in map(lambda base: self.visit(base, ctx), node.bases)))\
+                   .write_if(self.python_compatible, lambda: '''constructor(...args) {if ('__init__' in this) this.__init__(...args); return new Proxy(this, { apply: (target, self, args) => target.__call__(self, ...args)})}''')
+                   .write(self.visit(filter(lambda node: isinstance(node, ast.FunctionDef), body), context))
+                   .write(self.visit(filter(lambda node: not isinstance(node, ast.FunctionDef), body), context))
+                   )\
+            .write_if(node.bases, lambda: (f'''Object.getOwnPropertyNames({base}.prototype).forEach(name => {{if (name !== 'constructor') {{{node.name}.prototype[name] = {base}.prototype[name];}}}});''' for base in map(lambda base: self.visit(base, ctx), node.bases)))\
             .write(f'{node.name} = new Proxy({node.name}, {{ apply: (clazz, thisValue, args) => new clazz(...args) }});')\
             .write_if(node.decorator_list, (f'{node.name} = ', reduce(lambda value, element: f'{element}({value})', map(lambda decorator: self.visit(decorator, ctx), node.decorator_list), node.name), ';'))
 
@@ -150,17 +161,29 @@ class CodeGen(Visitor[JSVisitorContext]):
         context = ctx.copy(scope=JSScope.FUNCTION_FIELD)
 
         body = self.visit(node.body, context)
-
         if ctx.scope == JSScope.CLASS_FIELD:
-            return BaseBuilder()\
-                .write_if(context.has_yield, '*')\
-                .write(node.name)\
-                .write(f'{self.visit(node.args, ctx)}')\
-                .write(SurroundBuilder(surround='{}')
-                       .write(body))\
-                .build()
+            builder = BaseBuilder()
+            builder.write_if(context.has_yield, '*')
+            builder.write(node.name)
+            if self.python_compatible:
+                (
+                    builder
+                    .write('=(...__args)=>{')
+                    .write(f'({self.visit(node.args, ctx)} =>')
+                    .write(SurroundBuilder(surround='{}')
+                           .write(body))
+                    .write(')(this,...__args)}')
+                )
+            else:
+                (
+                    builder
+                    .write(self.visit(node.args, ctx))
+                    .write(SurroundBuilder(surround='{}')
+                           .write(body))
+                )
+            return builder.build()
         builder = BaseBuilder()\
-            .write_if(node.name not in ctx.variables, 'let')\
+            .write_if(not ctx.is_existed(node.name), 'let')\
             .write(node.name)\
             .write('=')\
             .write_if(context.has_yield, 'function*')\
@@ -185,7 +208,7 @@ class CodeGen(Visitor[JSVisitorContext]):
                        .write(body))\
                 .build()
         builder = BaseBuilder()\
-            .write_if(node.name not in ctx.variables, 'let')\
+            .write_if(not ctx.is_existed(node.name), 'let')\
             .write(node.name)\
             .write('=')\
             .write_if(context.has_yield, 'async function*', or_else='async')\
@@ -195,11 +218,11 @@ class CodeGen(Visitor[JSVisitorContext]):
                    .write(body))
         ctx.variables.append(node.name)
         return builder.build()
-    
+
     def visit_Yield(self, node: ast.Yield, ctx: JSVisitorContext):
         ctx.has_yield = True
         return f'yield {self.visit(node.value, ctx)}'
-    
+
     def visit_YieldFrom(self, node: ast.YieldFrom, ctx: JSVisitorContext):
         ctx.has_yield = True
         return f'yield* {self.visit(node.value, ctx)}'
@@ -212,7 +235,7 @@ class CodeGen(Visitor[JSVisitorContext]):
         invert = False
         for op in node.ops:
             op_type = type(op)
-            ops.write(ctx.compare_op_table[op_type])
+            ops.write(self.COMPARE_OP_TABLE[op_type])
             if op_type == ast.NotIn:
                 invert = True
 
@@ -229,21 +252,12 @@ class CodeGen(Visitor[JSVisitorContext]):
         return node.id
 
     def visit_Constant(self, node: ast.Constant, ctx: JSVisitorContext):
-        if node.value is None:
-            return 'null'
-        if isinstance(node.value, bool):
-            return 'true' if node.value else 'false'
-        if node.value in ctx.constant_table:
-            return ctx.constant_table[node.value]
-        if ctx.scope != JSScope.F_STRING and isinstance(node.value, str):
-            return f'`{node.value}`'
-        else:
-            return f'{node.value}'
+        return self.CONSTANT_CONSUMER_TABLE[type(node.value)](node.value, ctx)
 
     def visit_Assign(self, node: ast.Assign, ctx: JSVisitorContext):
         builder = BaseBuilder(separator=';')
 
-        defines = [target.id for target in node.targets if isinstance(target, ast.Name) and target.id not in ctx.variables]
+        defines = [target.id for target in node.targets if isinstance(target, ast.Name) and not ctx.is_existed(target.id)]
         ctx.variables.extend(defines)
 
         if len(node.targets) == len(defines):
@@ -287,9 +301,9 @@ class CodeGen(Visitor[JSVisitorContext]):
             .write_if(node.args, self.visit(args, context))
             .write_if(node.kwarg or node.defaults, lambda:
                       SurroundBuilder(surround='{}', separator=',')
-                      .write_if(keywords, lambda: (f'{self.visit(k, context)}={self.visit(v, context)}' for k, v in keywords.items()))
+                      .write_if(keywords, lambda: (f'{self.visit(key, context)}={self.visit(value, context)}' for key,  value in keywords.items()))
                       .write_if(node.kwarg, lambda: f'...{self.visit(node.kwarg, context)}')
-                      .build())
+                      .build() + '={}')
             .write_if(node.vararg, lambda: f'...{self.visit(node.vararg, context)}')
             .build()
         ).build()
@@ -324,24 +338,24 @@ class CodeGen(Visitor[JSVisitorContext]):
     def visit_For(self, node: ast.For, ctx: JSVisitorContext):
         target = BaseBuilder()
         if isinstance(node.target, ast.Name):
-            if node.target.id not in ctx.variables:
+            if not ctx.is_existed(node.target.id):
                 target.write(f'let')
         target.write(self.visit(node.target, ctx))
 
         return (
             BaseBuilder()
-            .write_if(node.orelse, '_else: {')
+            .write_if(node.orelse, '__else: {')
             .write(f'for ({target.build()} of {self.visit(node.iter, ctx)})')
-            .write(f"{{{self.visit(node.body, ctx.copy(break_suffix='_else' if node.orelse else ''))}}}")
+            .write(f"{{{self.visit(node.body, ctx.copy(break_suffix='__else' if node.orelse else ''))}}}")
             .write_if(node.orelse, f'{self.visit(node.orelse, ctx)} }}')
         ).build()
 
     def visit_AsyncFor(self, node: ast.AsyncFor, ctx: JSVisitorContext):
         return (
             BaseBuilder()
-            .write_if(node.orelse, '_else: {')
+            .write_if(node.orelse, '__else: {')
             .write(f'for await ({self.visit(node.target, ctx)} of {self.visit(node.iter, ctx)})')
-            .write(f"{{{self.visit(node.body, ctx.copy(break_suffix='_else' if node.orelse else ''))}}}")
+            .write(f"{{{self.visit(node.body, ctx.copy(break_suffix='__else' if node.orelse else ''))}}}")
             .write_if(node.orelse, f'{self.visit(node.orelse, ctx)} }}')
         ).build()
 
@@ -349,12 +363,13 @@ class CodeGen(Visitor[JSVisitorContext]):
         return f'await {self.visit(node.value, ctx)}'
 
     def visit_BoolOp(self, node: ast.BoolOp, ctx: JSVisitorContext):
-        return ctx.bool_op_table[type(node.op)].join(map(lambda value: self.visit(value, ctx), node.values))
+        return self.BOOL_OP_TABLE[type(node.op)].join(map(lambda value: self.visit(value, ctx), node.values))
 
     def visit_UnaryOp(self, node: ast.UnaryOp, ctx: JSVisitorContext):
-        return f'{ctx.unary_op_table[type(node.op)]}{self.visit(node.operand, ctx)}'
+        return f'{self.UNARY_OP_TABLE[type(node.op)]}{self.visit(node.operand, ctx)}'
 
     def visit_ListComp(self, node: ast.ListComp, ctx: JSVisitorContext):
+        ctx.parent = node
         return f'{self.visit(node.generators, ctx)}'
 
     def visit_DictComp(self, node: ast.DictComp, ctx: JSVisitorContext):
@@ -363,12 +378,13 @@ class CodeGen(Visitor[JSVisitorContext]):
 
     def visit_comprehension(self, node: ast.comprehension, ctx: JSVisitorContext):
         builder = BaseBuilder(f'{self.visit(node.iter, ctx)}')
-        builder.write_if(node.ifs, (f'.filter({self.visit(x, ctx)})' for x in node.ifs))
+        target = self.visit(node.target, ctx)
+        builder.write_if(node.ifs, (f'.filter({target}=>{self.visit(x, ctx)})' for x in node.ifs))
 
         if isinstance(ctx.parent, ast.DictComp):
-            builder.write(f'.map(({self.visit(node.target, ctx)})=>[{self.visit(ctx.parent.key, ctx)}, {self.visit(ctx.parent.value, ctx)}])')
+            builder.write(f'.map(({target})=>[{self.visit(ctx.parent.key, ctx)}, {self.visit(ctx.parent.value, ctx)}])')
         else:
-            builder.write(f'.map(({self.visit(node.target, ctx)})=>{self.visit(ctx.parent.elt, ctx)})')
+            builder.write(f'.map(({target})=>{self.visit(ctx.parent.elt, ctx)})')
 
         return builder.build()
 
@@ -379,12 +395,15 @@ class CodeGen(Visitor[JSVisitorContext]):
         return f'...{self.visit(node.value, ctx)}'
 
     def visit_BinOp(self, node: ast.BinOp, ctx: JSVisitorContext):
-        return f'{self.visit(node.left, ctx)}{ctx.operator_op_table[type(node.op)]}{self.visit(node.right, ctx)}'
+        op_type = type(node.op)
+        expr = f'{self.visit(node.left, ctx)}{self.OPERATOR_OP_TABLE[op_type]}{self.visit(node.right, ctx)}'
+        if op_type == ast.FloorDiv:
+            return f'(({expr}) >> 0)'
+        return f'({expr})'
 
     def visit_AnnAssign(self, node: ast.AnnAssign, ctx: JSVisitorContext):
         if isinstance(node.target, ast.Name):
             ctx.variables.append(node.target.id)
-        print(node.value is not None)
         return (
             BaseBuilder()
             .write_if(not isinstance(node.target, ast.Attribute) and ctx.scope != JSScope.CLASS_FIELD, 'let')
@@ -400,13 +419,13 @@ class CodeGen(Visitor[JSVisitorContext]):
         builder.write(
             BaseBuilder('let')
             .write(SurroundBuilder(surround='[]', separator=',')
-                   .write(f'_with_{i}' for i, expr in enumerate(node.items)))
+                   .write(f'__with_{i}' for i, expr in enumerate(node.items)))
             .write('=')
             .write(SurroundBuilder(surround='[]', separator=',').write(f'{self.visit(expr.context_expr)}' for expr in node.items))
         )\
-            .write(BaseBuilder(separator=';').write(BaseBuilder().write_if(expr.optional_vars, f'{self.visit(expr.optional_vars)}=').write(f'_with_{i}.__enter__()') for i, expr in enumerate(node.items)))\
+            .write(BaseBuilder(separator=';').write(BaseBuilder().write_if(expr.optional_vars, f'{self.visit(expr.optional_vars)}=').write(f'__with_{i}.__enter__()') for i, expr in enumerate(node.items)))\
             .write(self.visit(node.body))\
-            .write(BaseBuilder(separator=';').write(BaseBuilder().write_if(expr.optional_vars, f'{self.visit(expr.optional_vars)}=').write(f'_with_{i}.__exit__()') for i, expr in enumerate(node.items)))
+            .write(BaseBuilder(separator=';').write(BaseBuilder().write_if(expr.optional_vars, f'{self.visit(expr.optional_vars)}=').write(f'__with_{i}.__exit__()') for i, expr in enumerate(node.items)))
 
         return builder.build()
 
@@ -421,10 +440,23 @@ class CodeGen(Visitor[JSVisitorContext]):
         return ';'.join(map(lambda target: f'delete {self.visit(target, ctx)}', node.targets))
 
     def visit_AugAssign(self, node: ast.AugAssign, ctx: JSVisitorContext):
-        return f'{self.visit(node.target, ctx)} {ctx.operator_op_table[type(node.op)]}= {self.visit(node.value, ctx)}'
+        op_type = type(node.op)
+        target = self.visit(node.target, ctx)
+        expr = f'{target} {self.OPERATOR_OP_TABLE[type(node.op)]}= {self.visit(node.value, ctx)}'
+        if op_type == ast.FloorDiv:
+            return f'{expr};{target} = ({target}) >> 0'
+        return expr
 
     def visit_Subscript(self, node: ast.Subscript, ctx: JSVisitorContext):
-        return f'{self.visit(node.value, ctx)}[{self.visit(node.slice, ctx)}]'
+        slice = node.slice
+        if isinstance(slice, ast.Slice):
+            if slice.step is None:
+                return f'{self.visit(node.value, ctx)}.slice({self.visit(slice.lower, ctx)}, {self.visit(slice.upper, ctx)})'
+
+        return f'{self.visit(node.value, ctx)}[{self.visit(slice, ctx)}]'
+
+    def visit_Slice(self, node: ast.Slice, ctx: JSVisitorContext):
+        return f'slice({self.visit(node.lower, ctx)}, {self.visit(node.upper, ctx)}, {self.visit(node.step, ctx)})'
 
     def visit_While(self, node: ast.While, ctx: JSVisitorContext):
         return 'while (%s) {%s}' % (self.visit(node.test, ctx), self.visit(node.body, ctx))
@@ -444,15 +476,15 @@ class CodeGen(Visitor[JSVisitorContext]):
     def visit_Try(self, node: ast.Try, ctx: JSVisitorContext):
         return (
             BaseBuilder(separator='')
-            .write_if(node.orelse, '_else: {')
+            .write_if(node.orelse, '__else: {')
             .write(f'try {{{self.visit(node.body, ctx)}}}')
             .write_if(node.handlers, "catch(_err) {%s}" % '\n'.join(
                 BaseBuilder()
-                .write_if(handler.type, lambda: f'if (_err instanceof {self.visit(handler.type, ctx)})')
+                .write_if(handler.type, lambda: f'if (__err instanceof {self.visit(handler.type, ctx)})')
                 .write(SurroundBuilder(surround='{}', separator=';')
-                       .write_if(handler.name, f'{handler.name} = _err')
+                       .write_if(handler.name, f'{handler.name} = __err')
                        .write(self.visit(handler.body, ctx))
-                       .write_if(node.orelse, 'break _else')
+                       .write_if(node.orelse, 'break __else')
                        )
                 .build()
                 for handler in node.handlers
@@ -464,11 +496,15 @@ class CodeGen(Visitor[JSVisitorContext]):
     def visit_Match(self, node: ast.Match, ctx: JSVisitorContext):
         return '/*@deprecated match not supported */'
 
+    def visit_Global(self, node: ast.Global, ctx: JSVisitorContext):
+        ctx.global_variables.extend(node.names)
+        return f'var {",".join(node.names)}'
+
 
 space_regex = re.compile('\s+')
 
 
-def js(func: t.Optional[t.Callable] = None, as_function: bool = False, python_compatible: bool = False) -> str:
+def js(func: Optional[Callable] = None, as_function: bool = False, python_compatible: bool = False) -> str:
     def wrapper(func):
         lines = inspect.getsourcelines(func)[0]
 
@@ -500,9 +536,9 @@ def js(func: t.Optional[t.Callable] = None, as_function: bool = False, python_co
 
 
 def convert(
-    source: t.Union[str, ast.AST, types.ModuleType, types.FunctionType],
-    formatter: t.Optional[t.Callable[[str], str]] = jsbeautifier.beautify,
-    code_gen: t.Optional[Visitor] = None,
+    source: Union[str, ast.AST, types.ModuleType, types.FunctionType],
+    formatter: Optional[Callable[[str], str]] = jsbeautifier.beautify,
+    code_gen: Optional[Visitor] = None,
     python_compatible: bool = False
 ):
     generator = code_gen or CodeGen(python_compatible=python_compatible)
